// grad.cc
// version 1.0, SM, May 2021
//              based on BuHa by Lg
// version 1.1, SM, May 2021
// version 1.2, SM, May 2021
// version 1.3, SM, May 2021
// version 1.4, SM, May 2021
// version 1.5, SM, June 2021
// version 1.6, SM, June 2021
// version 1.7, SM, June 2021
// version 1.8, SM, July 2021
// version 1.9, SM, July 2021
// version 1.10, SM, August 2021
// version 1.11, SM, May 2022
// version 1.12, SM, July 2022
// version 1.13, SM, December 2023

#include <iostream>
#include <fstream>
#include <string>
#include "uti1a.h"
#include "lineInput.h"
#include "menu.h"
#include "grad.h"

const char* mid = "grad (December 2023, v1.13)";
const char* prompt = "grad> ";
const char* defaultfilename = "default";

long int line_count=-1;
menu men( &line_count );

istream* inp = &cin;
ostream* outp = &cout;

int fileinput=0;     // boolean
int _terminate=0;     // boolean

enum { isExam, isStudent, doesNotExist, noAccount };

const int bufsize = 256;
char foldername[bufsize] = "default_LaTeX_folder";
char classname[bufsize] = "";

static void quit() { _terminate=1; }

static void skipline() { li_skipline(*inp); }

static void syntax_error()
{
    input_error( line_count, "syntax error" );
    skipline();
}

//////////////////////////////////////////////////////////////////////////

exam_list exams;
student_list students;
float rounding=0.25;
float maxgrade=6;
float mingrade=1;

//////////////////////////////////////////////////////////////////////////

static void set_rounding()
{
    float r;
    if ( !li_get_float( *inp, r ) ) {
        input_error( line_count, "error: number expected" );
        skipline(); return; 
    }
    rounding = r;
    skipline();
    //*outp << endl << "Test: rounding=" << r << endl;
}


static void set_range()
{
  float rmin;
  if ( !li_get_float( *inp, rmin ) ) {
    input_error( line_count, "error: number expected" );
    skipline(); return; 
  }
  float rmax;
  if ( !li_get_float( *inp, rmax ) ) {
    input_error( line_count, "error: number expected" );
    skipline(); return; 
  }
  if (rmax <= rmin) {
    input_error( line_count, "error: maximum must be larger than minimum" );
    skipline(); return; 
  }
  maxgrade = rmax;
  mingrade = rmin;
  skipline();
  //*outp << endl << "Test: rmin=" << rmin << "  rmax=" << rmax << endl;
}



//////////////////////////////////////////////////////////////////////////

static int is_legal_in_filename( char ch )
{
    return isalpha(ch) || isdigit(ch)
           || (ch=='/')
           || (ch=='_')
           || (ch=='.')
           || (ch==':')
           || (ch=='~');
}


static int is_legal_in_account_name( char ch )
{
    return isalpha(ch) || isdigit(ch) 
           || (ch=='/')
           || (ch=='_')
           || (ch=='.');
}


static int is_legal_in_comment( char ch )
{
    return ch!='#';
}


///////////////////////////////////////////////////////////////////////////

class infile_access 
{ //  -------------
  private:
    ifstream *inp;
    long int line_count;
    infile_access *prev;
  friend void infile();
  friend void defaultinfile();
  friend void closefile();
};

int nesting_depth=0;
infile_access *fap=0;
ifstream *cinp=0;   // current file

void infile()
{
    const int max_files = 10;    // change freely, just here to 
                                 // prevent accidental recursion
    if ( nesting_depth >= max_files ) {
        input_error( line_count, "error: nested file input beyond maximum depth" );
        skipline(); return;
    }
    li_eatspace(*inp);
    char filename[bufsize];
    int nchar = 
    li_get_string(*inp, filename, bufsize, &is_legal_in_filename, (int)line_count); 
    if ( nchar >= bufsize ) {
      input_error( line_count, "buffer error");
      skipline(); return;
    }
    if ( nchar==0 ) { input_error( line_count, "error: no filename" ); return; }
    ifstream *_inp = new ifstream(filename);
    if ( _inp==0 ) memory_error();  // might depend on C++ implementation
    if ( !_inp->is_open() ) {
        input_error( line_count, "error: can't open ", filename );
        delete _inp;
        return;
    }
    infile_access *_fap = fap;
    fap = new infile_access; 
    if ( fap==0 ) memory_error();
    //
    //*outp << "opening file " << filename 
    //      << " at level " << nesting_depth << endl;
    //
    fap->inp = cinp;
    fap->line_count = line_count;
    fap->prev = _fap;
    inp = cinp = _inp;
    line_count=1;
    fileinput=1;
    nesting_depth++;
}


void defaultinfile()
{
    const int max_files = 10;    // change freely, just here to 
                                 // prevent accidental recursion
    if ( nesting_depth >= max_files ) {
        input_error( line_count, "error: nested file input beyond maximum depth" );
        skipline(); return;
    }
    li_eatspace(*inp);
    ifstream *_inp = new ifstream(defaultfilename);
    if ( _inp==0 ) memory_error();  // might depend on C++ implementation
    if ( !_inp->is_open() ) {
        input_error( line_count, "error: can't open ", defaultfilename );
        delete _inp;
        return;
    }
    infile_access *_fap = fap;
    fap = new infile_access; 
    if ( fap==0 ) memory_error();
    fap->inp = cinp;
    fap->line_count = line_count;
    fap->prev = _fap;
    inp = cinp = _inp;
    line_count=1;
    fileinput=1;
    nesting_depth++;
}


void closefile()
{
    if ( fap==0 ) return;    // no file open
    delete cinp;
    cinp = fap->inp;
    inp = (cinp==0) ? &cin : cinp;
    line_count = fap->line_count;
    infile_access *prev = fap->prev;
    delete fap;
    fap = prev;
    nesting_depth--;
    if ( nesting_depth==0 ) fileinput=0;
    //
    //*outp << "file closed, continueing at level " << nesting_depth;
    //if ( nesting_depth>0 ) *outp << " at line " << line_count;
    //*outp << endl;
}

////////////////////////////////////////////////////////////////////////

static student* get_student( int no_account_is_error=0 )
{
  char buf[bufsize];
  li_eatspace(*inp);
  int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
  if ( nchars >= bufsize ) {
    input_error( line_count, "buffer error");
    skipline(); return 0;
  }
  if ( nchars==0 ) {
    if ( no_account_is_error ) {
      input_error( line_count, "error: student expected" ); 
      skipline();
    }
    return 0;
  }
  int status;
  student* sp = (student*) students.search( buf, &status );
  if ( status != simpleStringList::exact_match ) { 
    input_error( line_count, "error: unknown student ", buf );
    skipline(); return 0; 
  }
  return sp;
}


static exam* get_exam( int no_account_is_error=0 )
{
  char buf[bufsize];
  li_eatspace(*inp);
  int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
  if ( nchars >= bufsize ) {
    input_error( line_count, "buffer error");
    skipline(); return 0;
  }
  if ( nchars==0 ) {
    if ( no_account_is_error ) {
      input_error( line_count, "error: exam/grade expected" ); 
      skipline();
    }
    return 0;
  }
  int status;
  exam* ep = (exam*) exams.retrieve( buf, &status );
  if ( status != simpleStringList::exact_match ) { 
    input_error( line_count, "error: unknown exam/grade ", buf );
    skipline(); return 0; 
  }
  return ep;
}


static simpleStringListEntry* get_account( int* statusp )
{
  char buf[bufsize];
  li_eatspace(*inp);
  int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
  if ( nchars >= bufsize )  {
    input_error( line_count, "buffer error");
    skipline(); return 0;
  }
  if ( nchars==0 ) {
    *statusp = noAccount;
    return 0;
  }
  int status;
  simpleStringListEntry* sep = exams.retrieve( buf, &status );
  if ( status == simpleStringList::exact_match ) { 
    *statusp = isExam;
    return sep;
  }
  sep = students.search( buf, &status );
  if ( status == simpleStringList::exact_match ) { 
    *statusp = isStudent;
    return sep;
  }
  *statusp = doesNotExist;
  input_error( line_count, "error: unknown exam or student ", buf );
  skipline(); return 0;
}

//////////////////////////////////////////////////////////////////////////

static void booking() 
{
  exam* examp = get_exam();
  if ( examp==0 ) return;
  li_eatspace(*inp);
  char ch;
  if ( !li_get(*inp,ch) || ((ch!='-') && (ch!='+')) ) { 
    input_error( line_count, "syntax error: missing \"-\" or \"+\"" ); 
    skipline(); return;
  }
  student* studp = get_student(1);
  if ( studp==0 ) return; 
  if (ch=='+') {
    if ( examp->isexam() ) {
      input_error( line_count, "error: exam grades cannot be overwritten: ", examp->strngp );
      skipline(); return;
    }
    exam_entry* eep=examp->stud_in_exam_search(studp->strngp);
    if ( !eep ) {
      input_error( line_count, "error: student does not have a grade in the specified grade, so the grade cannot be overwritten: ", studp->strngp );
      skipline(); return;
    }
    float grade;
    if ( li_get_float( *inp, grade ) ) {
      examp->overwrite_gradeentry( eep, grade );
    } else {
      input_error( line_count, "syntax error: missing grade" ); 
      skipline(); return;
    }
  } else {
    if ( examp->stud_in_exam_search(studp->strngp) ) {
      input_error( line_count, "error: student already has a grade in the specified exam: ", studp->strngp );
      skipline(); return;
    }
    if ( examp->isexam() ) {
      pointslist *pointsp = new pointslist();
      if ( pointsp==0 ) memory_error();
      float point;
      if ( li_get_float( *inp, point ) ) {
	pointsp->put_entry(point);
      } else {
	input_error( line_count, "error: missing pointlist" ); 
	skipline(); return;
      }
      for (;;) {
	if ( li_get_float( *inp, point ) ) {
	  pointsp->put_entry(point);
	} else {
	  skipline();
	  break;
	}
      }
      if ( examp->put_examentry(studp, pointsp, rounding, maxgrade, mingrade) ) {
	input_warning( line_count, "warning: number of problems not consistent" );
      }
    } else {
      float grade;
      if ( li_get_float( *inp, grade ) ) {
	examp->put_gradeentry( studp, grade );
      } else {
	input_error( line_count, "syntax error: missing grade" ); 
	skipline(); return;
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////

static void avg() 
{
  exam* to = get_exam(1); 
  if ( to==0 ) return; 
  if ( to->isexam() ) {
    input_error( line_count, "error: grade expected, not exam: ", to->strngp ); 
    skipline(); return;
  }
  if ( to->listLength() ) {
    input_error( line_count, "error: grade not empty: ", to->strngp ); 
    skipline(); return;
  }
  li_eatspace(*inp);
  char ch;
  if ( !li_get(*inp,ch) || (ch!='<') ) { 
    input_error( line_count, "syntax error: missing \"<\"" ); 
    skipline(); return;
  }
  //first exam in list
  exam* from = get_exam(1);
  if ( from==0 ) {
    input_error( line_count, "error: exam/grade expected" );
    skipline(); return;
  }
  if ( from==to ) { 
    input_error( line_count, "error: exams must be different", from->strngp );
    skipline(); return;
  }
  to->add_avg(from);
  //remaining exams in list
  while (from) {
    from = get_exam(0);
    if ( from==0 ) break; 
    if ( from==to ) { 
      input_error( line_count, "error: exams must be different", from->strngp );
      skipline(); return;
    }
    to->add_avg(from);
  }
  //compute average
  to->add_avg(from,rounding,maxgrade,mingrade);
  //to->sort(sorting);
  skipline();
}


static void max() 
{
  exam* to = get_exam(1); 
  if ( to==0 ) return; 
  if ( to->isexam() ) {
    input_error( line_count, "error: grade expected, not exam: ", to->strngp ); 
    skipline(); return;
  }
  if ( to->listLength() ) {
    input_error( line_count, "error: grade not empty: ", to->strngp ); 
    skipline(); return;
  }
  li_eatspace(*inp);
  char ch;
  if ( !li_get(*inp,ch) || (ch!='<') ) { 
    input_error( line_count, "syntax error: missing \"<\"" ); 
    skipline(); return;
  }
  //first exam in list
  exam* from = get_exam(1);
  if ( from==0 ) {
    input_error( line_count, "error: exam/grade expected" );
    skipline(); return;
  }
  if ( from==to ) { 
    input_error( line_count, "error: exams must be different", from->strngp );
    skipline(); return;
  }
  to->add_max(from);
  //remaining exams in list
  while (from) {
    from = get_exam(0);
    if ( from==0 ) break; 
    if ( from==to ) { 
      input_error( line_count, "error: exams must be different", from->strngp );
      skipline(); return;
    }
    to->add_max(from);
  }
  //rounding
  to->add_max(from,rounding,maxgrade,mingrade);
  //to->sort(sorting);
  skipline();
}


static void min() 
{
  exam* to = get_exam(1); 
  if ( to==0 ) return; 
  if ( to->isexam() ) {
    input_error( line_count, "error: grade expected, not exam: ", to->strngp ); 
    skipline(); return;
  }
  if ( to->listLength() ) {
    input_error( line_count, "error: grade not empty: ", to->strngp ); 
    skipline(); return;
  }
  li_eatspace(*inp);
  char ch;
  if ( !li_get(*inp,ch) || (ch!='<') ) { 
    input_error( line_count, "syntax error: missing \"<\"" ); 
    skipline(); return;
  }
  //first exam in list
  exam* from = get_exam(1);
  if ( from==0 ) {
    input_error( line_count, "error: exam/grade expected" );
    skipline(); return;
  }
  if ( from==to ) { 
    input_error( line_count, "error: exams must be different", from->strngp );
    skipline(); return;
  }
  to->add_min(from);
  //remaining exams in list
  while (from) {
    from = get_exam(0);
    if ( from==0 ) break; 
    if ( from==to ) { 
      input_error( line_count, "error: exams must be different", from->strngp );
      skipline(); return;
    }
    to->add_min(from);
  }
  //rounding
  to->add_min(from,rounding,maxgrade,mingrade);
  //to->sort(sorting);
  skipline();
}


static void sort() 
{
  li_eatspace(*inp);
  char ch;
  char space;
  int order;
  if ( !li_get(*inp,ch) || ((ch!='n') && (ch!='g') && (ch!='c'))
       || ( li_get(*inp,space) && (space!='\t') && (space!=' '))) { 
    input_error( line_count, "syntax error: \"n\", \"g\", or \"c\" expected" ); 
    skipline(); return;
  }
  if (ch=='g') {
    order=exam::sortbygrade;
  } else if (ch=='c') {
    order=exam::sortbycomment;
  } else {
    order=exam::sortbyname;
  }
  exam* g = get_exam(0);
  if ( g==0 ) {
    if ( order != exam::sortbygrade ) {
      students.sort(order);
    }
  }
  while (g) {
    g->sort(order);
    g = get_exam(0);
  }
  skipline();
}


//////////////////////////////////////////////////////////////////////

static void define_student()
{
    char buf[bufsize];
    li_eatspace(*inp);
    int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
    if ( nchars >= bufsize )  {
      input_error( line_count, "buffer error");
      skipline(); return;
    }
    if ( nchars==0 ) {
        input_error( line_count, "student name expected" );
        skipline(); return; 
    }
    int status;
    simpleStringListEntry* slep;
    slep = exams.retrieve( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "name already used by exam/grade: ", buf );
        skipline(); return;
    }
    slep = students.search( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "redefinition of student ", buf );
        skipline(); return;
    }
    char* name = new char[strlen(buf)+1];
    if ( name==0 ) memory_error();
    strcpy( name, buf );
    li_eatspace(*inp);
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
    student *studp = new student(name, buf);
    if ( studp==0 ) memory_error();
    students.append( studp );
}


static void define_exam()
{
    char buf[bufsize];
    li_eatspace(*inp);
    // read exam name
    int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
    if ( nchars >= bufsize )  {
      input_error( line_count, "buffer error");
      skipline(); return;
    }
    if ( nchars==0 ) {
        input_error( line_count, "exam name expected" );
        skipline(); return; 
    }
    int status;
    simpleStringListEntry* slep;
    slep = students.search( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "name already used by student: ", buf );
        skipline(); return;
    }
    slep = exams.retrieve( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "redefinition of exam/grade ", buf );
	skipline(); return;
    }
    char* name = new char[strlen(buf)+1];
    if ( name==0 ) memory_error();
    strcpy( name, buf );
    // read weight
    float weight;
    if ( !li_get_float( *inp, weight ) ) {
        input_error( line_count, "weight expected" );
        skipline(); return; 
    }
    // read points for linear function
    float x1;
    if ( !li_get_float( *inp, x1 ) ) {
        input_error( line_count, "Point for linear function expected" );
        skipline(); return; 
    }
    float y1;
    if ( !li_get_float( *inp, y1 ) ) {
        input_error( line_count, "Point for linear function expected" );
        skipline(); return; 
    }
    float x2;
    if ( !li_get_float( *inp, x2 ) ) {
        input_error( line_count, "Point for linear function expected" );
        skipline(); return; 
    }
    float y2;
    if ( !li_get_float( *inp, y2 ) ) {
        input_error( line_count, "Point for linear function expected" );
        skipline(); return; 
    }
    // read comment
    li_eatspace(*inp);
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
    // create exam
    exam *examp = new exam(name, buf, 1, weight, x1, y1, x2, y2);
    if ( examp==0 ) memory_error();
    exams.insert( examp, slep );
}


static void define_grade()
{
    char buf[bufsize];
    li_eatspace(*inp);
    // read exam name
    int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
    if ( nchars >= bufsize )  {
      input_error( line_count, "buffer error");
      skipline(); return;
    }
    if ( nchars==0 ) {
        input_error( line_count, "exam name expected" );
        skipline(); return; 
    }
    int status;
    simpleStringListEntry* slep;
    slep = students.search( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "name already used by student: ", buf );
        skipline(); return;
    }
    slep = exams.retrieve( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "redefinition of exam/grade ", buf );
        skipline(); return;
    }
    char* name = new char[strlen(buf)+1];
    if ( name==0 ) memory_error();
    strcpy( name, buf );
    // read weight
    float weight;
    if ( !li_get_float( *inp, weight ) ) {
        input_error( line_count, "weight expected" );
        skipline(); return; 
    }
    // read comment
    li_eatspace(*inp);
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
    // create exam
    exam *examp = new exam(name, buf, 0, weight);
    if ( examp==0 ) memory_error();
    exams.insert( examp, slep );
}

//////////////////////////////////////////////////////////////////////

static void reset()
{
  exams.clear(1);
  students.clear(1);
}

/*
static void clear()
{
  exam* examp = get_exam(1);
  while (examp) {
    examp->clear();
    examp = get_exam(0);
  }
  skipline();  
}
*/


static void hide()
{
  exam* ep = get_exam(1);
  if ( ep==0 ) {
    input_error( line_count, "error: exam/grade expected" );
    skipline(); return;
  }
  while (ep) {
    ep->togglehidden();
    ep = get_exam(0);
  }
  skipline();
}


////////////////////////////////////////////////////////////////////

static void list()
{
    students.print_student_list(*outp);
    *outp << "\n";
    exams.print_exam_list(*outp);
    skipline();
}  

static void show()
{
  int status;
  simpleStringListEntry* sep = get_account(&status);
  if ( status == noAccount ) {
    students.print_overview(*outp, &exams);
    skipline();
  } else if ( status == isExam ) {
    ((exam*)sep)->print(*outp);
    skipline();
  } else if ( status == isStudent ) {
    ((student*)sep)->print(*outp);
    skipline();
  } 
}


static void line() 
{ 
  *outp << "\n";
}


static void stop() 
{ 
  //*outp << "Gugus";
  closefile();
}


static void setLaTeXfolder()
{
  li_eatspace(*inp);
  int nchar = 
    li_get_string(*inp, foldername, bufsize, &is_legal_in_filename, (int)line_count); 
  if ( nchar >= bufsize )  {
    input_error( line_count, "buffer error");
    skipline(); return;
  }
  if ( nchar==0 ) { input_error( line_count, "no folder" ); return; }
  //*outp << "setting folder name to " << foldername << " (note that it is assumed that this folder exists!)" << endl;
}


static void setclass()
{
  li_eatspace(*inp);
  int nchar = 
    li_get_string(*inp, classname, bufsize, &is_legal_in_comment, (int)line_count); 
  if ( nchar >= bufsize )  {
    input_error( line_count, "buffer error");
    skipline(); return;
  }
  if ( nchar==0 ) {
    *outp << "class name erased" << endl;
    return;
  }
  *outp << endl << "class: " << classname << endl;
}


static void printAllLaTeX()
{
  li_eatspace(*inp);
  int naming=0; // 0=name, 1=comment, 2=anonymous
  int landscape=0; //0=portraint, 1=landscape
  int no_of_fields=-1; //number of fields in empty stud list
  char ch;
  *outp << "writing files to (presumably) existing folder " << foldername << "..." << endl;
  while (li_get(*inp,ch)) {
    if (isdigit(ch)) {
      inp->putback(ch);
      if (no_of_fields!=-1) {
	input_error( line_count, "syntax error: two numbers given as optional argument");
	skipline(); return;
      }
      li_get_unumber(*inp,no_of_fields);
    } else if (ch == 'c') {
      if (naming != 0) {
	input_error( line_count, "syntax error: optional arguments 'c' and 'a' are exclusive and can only be given once");
	skipline(); return;
      }
      naming=1;
    } else if (ch == 'a') {
      if (naming != 0) {
	input_error( line_count, "syntax error: optional arguments 'c' and 'a' are exclusive and can only be given once");
	skipline(); return;
      }
      naming=2;
    } else if (ch == 'l') {
      if (landscape != 0) {
	input_error( line_count, "syntax error: optional argument 'l' can only be given once");
	skipline(); return;
      }
      landscape=1;
    }
  }
  if (no_of_fields == -1) {
    no_of_fields=12;
  }
  exams.print_latex_tofiles(foldername, naming, landscape, classname);
  students.print_latex_tofiles(foldername, naming, landscape, classname, no_of_fields, &exams);
}


static void help() 
{
    *outp << endl << mid << endl 
          << "available commands:\n\n"; 
    men.print_command_list(*outp, 1, 19); 
    *outp << endl;
    //*outp << "see also \\list\n";
}


static void init_menu()
{
    men.def( "\\", &syntax_error, "", 1 );
    men.def( "\\help", &help, "print this list of commands\n" );
    //
    men.def( "\\student", &define_student, 
             "define new student\n\t\t  grad> \\student name comments\n" );
    men.def( "\\exam", &define_exam, 
             "define new exam:\n\t\t  grad> \\exam name weight x1 y1 x2 y2 comments\n\t\tThe weight will be used in the computation of averages (see\n\t\t\\avg). (x1,y1) and (x2,y2) are two points that define the\n\t\tlinear scale used for converting points into grades. The linear\n\t\tscale is cut off at minimum and maximum defined by \\range and\n\t\tthe grade is rounded according to precision set by \\rounding.\n" );
    men.def( "\\grade", &define_grade, 
             "define new grade:\n\t\t  grad> \\grade name weight comments\n\t\tThe weight will be used in the computation of averages (see\n\t\t\\avg).\n" );
    //
    men.def( "\\sort", &sort,
	     "sort exams and grades:\n\t\t\tn sort according student name\n\t\t\tg sort according to grade\n\t\t\tc sort according to student's comment\n\t\tIf no exams or grades are given, the student list is sorted\n\t\tinstead (unless the sorting order is 'g', which is then\n\t\tignored).\n\t\t  grad> \\sort n s1 s2\n\t\t  grad> \\sort c\n" );
    men.def( "\\avg", &avg,
	     "compute resulting average grade from list of given exams and\n\t\tgrades, taking into account the weight of each exam and grade.\n\t\tThe result is rounded according to precision set by \\rounding\n\t\tand cut off at minimum and maximum defined by \\range.\n\t\t  grad> \\avg final < s1 s2 s3\n" );
    men.def( "\\max", &max,
	     "pick maximum grade from list of given exams and grades, rounded\n\t\taccording to precision set by \\rounding and cut off at minimum\n\t\tand maximum defined by \\range.\n\t\t  grad> \\max final < s1 s2 s3\n" );
    men.def( "\\min", &min,
	     "pick minimum grade from list of given exams and grades, rounded\n\t\taccording to precision set by \\rounding, and cut off at minimum\n\t\tand maximum defined by \\range.\n\t\t  grad> \\min final < s1 s2 s3\n" );
    men.def( "\\range", &set_range,
	     "set minimum and maximum value for grades. Default is 1 6.\n\t\t  grad> \\range 1 6\n" );
    men.def( "\\rounding", &set_rounding,
	     "set rounding precision for computing grades. If set to 0,\n\t\trounding is switched off. Default is 0.25.\n\t\t  grad> \\rounding 0.5\n" );
    //
    men.def( "\\\\", &defaultinfile,
	     "read file 'default', i.e., equivalent to\n\t\t  grad> \\infile default\n" );
    men.def( "\\infile", &infile,
	     "read file\n\t\t  grad> \\infile myfile\n" );
    men.def( "\\stop", &stop, "ignore rest of file\n" );
    //
    men.def( "\\line", &line, "print empty line\n" );
    //
    men.def( "\\list", &list,
	     "list of all students, exams, and grades\n" );
    men.def( "\\show", &show,
	     "with student or exam/grade: show student or exam/grade\n\t\twithout argument: show overview of all students and all\n\t\ttheir exam grades and grades\n\t\t  grad> \\show s1\n\t\t  grad> \\show laura\n\t\t  grad> \\show\n" );    
    //
    men.def( "\\setlatexfolder", &setLaTeXfolder,
	     "set path and name of folder where all LaTeX files will be\n\t\twritten to. Default is 'default_LaTeX_folder'.\n\t\t  grad> \\setlatexfolder myfolder\n");
    men.def( "\\class", &setclass,
	     "set name of class, used as title on all LaTeX files. If no\n\t\targuments are given, the class name is reset to the empty\n\t\tstring. Default is empty.\n\t\t  grad> \\class name used as title\n");
    men.def( "\\printlatex", &printAllLaTeX,
	     "print all students, exams, and grades, each as a separate\n\t\tLaTeX file, into the folder previously set by \\setlatexfolder.\n\t\tIn addition the files 'z_overview.tex', which shows an\n\t\tan overview of all students and their grades, and\n\t\t'z_studlist.tex', which is an empty table of all students,\n\t\tare generated.\n\t\tNB: To compile these LaTeX-files the stylefile 'research18.sty'\n\t\tis required. This stylefile is distributed together with the\n\t\tsource code of this program.\n\t\tOptional arguments:\n\t\t\tl\t landscape format\n\t\t\ta\t anonymous: no student names are printed\n\t\t\tc\t comment instead of name is printed\n\t\t\tinteger\t number of columns in empty student table\n\t\t\t\t (default is 12)\n\t\t  grad> \\printlatex cl14\n\t\t  grad> \\printlatex l\n");
    //
    men.def( "\\hide", &hide,
	     "hide (or unhide if already hidden) exams/grades: a hidden\n\t\tgrade/exam does not show up in overview or in printlatex\n\t\t  grad> \\hide s1 s2\n" );
    men.def( "\\reset", &reset, "erase everything\n" );
    men.def( "\\quit", &quit, "exit program\n" );
}

//////////////////////////////////////////////////////////////////////

int main()
{
    init_menu();
    char ch;
    *outp << mid << " (type \"\\help\" for help)\n" << prompt;
    while ( *inp && *outp && !_terminate ) {
        li_eatspace(*inp);
        if ( inp->get(ch)==0 ) { 
            closefile();
            continue;
        }
        if ( (ch=='\n') || (ch=='\r') ) {
            if (fileinput) line_count++;
            else *outp << prompt;
            continue;
        }
        inp->putback(ch);
        if ( ch=='\\' ) men.get_and_execute_command(*inp);
        else if ( is_legal_in_account_name(ch) ) booking();
        else {
            if ( ch!='%' ) input_error( line_count, "illegal character" );
            skipline();
        }
    }
    return 0;
}
