// grad.h
// version 1.0, SM, May 2021
//              based on BuHa by Lg
// version 1.1, SM, May 2021
// version 1.2, SM, May 2021
// version 1.3, SM, May 2021
// version 1.4, SM, May 2021
// version 1.5, SM, June 2021
// version 1.7, SM, June 2021
// version 1.8, SM, July 2021
// version 1.9, SM, July 2021

#ifndef GRAD_H
#define GRAD_H

#include "simpleList.h"

class student;     // forward
class exam;        // forward


class points_entry: public simpleListEntry {
  //  ============
  friend class pointslist;
  friend class exam;
protected:
  float _points;
};


class pointslist: public simpleList {
  //  ===========
  friend class exam;
public:
  float totalpoints;
  pointslist() { totalpoints=0; }
  int number_of_problems() { return listLength(); }
  void put_entry( float points );
  void clear();    // delete whole list
  ~pointslist() { clear(); }
};


class computegrade_entry: public simpleListEntry {
  //  ==================
  friend class computegradelist;
  friend class exam;
protected:
  exam* _gradep;
};


class computegradelist: public simpleList {
  //  ================
  friend class exam;
protected:
  int _type;
public:
  enum { isavg, ismax, ismin };
  computegradelist( int type ) { _type = type; }
  void put_entry( exam* ep );
  void clear();    // delete whole list
  ~computegradelist() { clear(); }
};


class exam_entry: public simpleListEntry {
  //  ===========
  friend class exam;
  friend class exam_list;
  friend class student;
protected:
  student* _studp;
  exam* _examp;
  pointslist* _pointsp;
  float _exactgrade; 
  float _grade;
  float _tmpweight;
};


class exam: public simpleList, public simpleStringListEntry {
  //  =====
  friend class student;
protected:
  int _isexam;
  int _hiddenflag;
  float _weight;
  float _x1, _y1, _x2, _y2;
  int _numberofproblems;
  float _exactsum;
  float _exactsquaredsum;
  float _sum;
  float _squaredsum;
  pointslist* _pointssump;
  computegradelist* _cglp;
  char* _comment;
public:
  exam(char* name, char* comment, int isexam, float weight,
       float x1=0, float y1=0, float x2=0, float y2=0);
  int isexam() { return _isexam; }
  int ishidden() { return _hiddenflag; }
  void togglehidden() { _hiddenflag = 1 - _hiddenflag; }
  float weight() { return _weight; }
  char* comment() { return _comment; }
  float average() { return _sum/listLength(); }
  void put_gradeentry( student* studp, float grade ); 
  void overwrite_gradeentry( exam_entry* eep, float grade );
  int put_examentry( student* studp, pointslist* pointsp, float rounding, float maxgrade, float mingrade ); //returns 1 if number of problems is inconsistent, but adds anyway
  exam_entry* stud_in_exam_search(const char* studname); // returns exam_entry if studname is found in list, 0 otherwise
  enum { sortbygrade, sortbyname, sortbycomment };   
  void sort(int order);
  int stringcomp(const char* first, const char* second);
  void add_avg( exam* from, float rounding=-1, float maxgrade=0, float mingrade=0 );
  void add_max( exam* from, float rounding=-1, float maxgrade=0, float mingrade=0 );
  void add_min( exam* from, float rounding=-1, float maxgrade=0, float mingrade=0 );
  void print( ostream&, int extraspace=0 );
  void print_LaTeX( ostream& os, int naming, int landscape, char* classname ); 
  void clear_comment();
  void clear();    // delete whole list
  ~exam() { clear_comment(); clear(); }
};


class exam_list: public simpleStringList {
  //  ==========
public:
  void print_exam_list( ostream& os, int tab=15, int width=10 ); // print only names, weight and gradetype
  void print_exams_for_student( ostream& os, const char* st, int width);
  int print_exams_for_student_latex( ostream& os, const char* st, int mode );
  void print_latex_tofiles( const char* folder, int naming, int landscape, char* classname ); // print LaTeX files
  void clear( int delete_account_names=0 );     // delete whole list
  ~exam_list() { clear(); }
};


class student_entry: public simpleListEntry {
  //  ==============
public:
  exam_entry* eep;
};


class student: public simpleList, public simpleStringListEntry {
  //  ========
  friend class exam;
protected:
  char* _comment;
  char* _indexcomment; // identical to _comment, but with accents removed and alll lowercase -> for sorting 
public:
  student( char* name, char* comment=0 ); //comment is copied
  void print( ostream&, int extraspace=0 );
  char* comment() { return _comment; }
  char* index() { return _indexcomment; }
  void print_LaTeX( ostream& os, int landscape, char* classname ); 
  void clear_comment();
  void clear();   // delete whole list of student_entries 
  ~student() { clear_comment(); clear(); }   // delete whole list and comment, but not student name
};


class student_list: public simpleStringList {
  //  =============
public:
  void sort(int order);
  int stringcomp(const char* first, const char* second);
  void print_student_list( ostream& os, int tab=15, int width=10 ); // print only names
  void print_overview( ostream& os, exam_list* elp, int width=6 ); // print all students with grades
  void print_latex_tofiles( const char* folder, int naming, int landscape, char* classname, int no_of_fields, exam_list* elp ); // print LaTeX files
  void clear( int delete_account_names=0 );  // delete whole list of students
  ~student_list() { clear(); }
};




#endif
