// lineInput.h          Lg, July 19, 1995
// version 2.0          SM, May 2021

#ifndef lineInput_H
#define lineInput_H


#include <iostream>
using namespace std;


int li_get( istream& is, char& c );
/*  ======  Does not read (i.e., puts back) '\n' and '\r'. 
            Returns 1 if a character was read and 0 otherwise;
            in the latter case, c is not changed.
            */

int li_eatspace( istream& is );
/*  ===========  Skip to next char different from ' ' and '\t';
			Returns 1 if at least one such character was read, 
			else return 0.
			*/

inline
void li_skipline( istream& is ) { char ch; while(li_get(is,ch)){;} }
//   ===========


                        /* (definition below)
template <class number>
int  li_get_unumber( istream& is, number& x );
     ==============  Read unsigned decimal number (using "li_get"). 
			Returns number of digits (leading zeros included). 
			Number x is not changed if 0 is returned. 
			OBS: overflow check not yet implemented.
			*/

int  li_get_float( istream& is, float& x );
/*   ============  Read floating number (using "li_get"). 
			Returns 0 if no number is read, 1 if an integer is
			read, and 2 if a fraction is read.
			*/


int li_get_string( istream& is, char* buf, int bufsize, 
/*  ============= */
                  int (*is_allowed)(char), // pointer to function
		  int line_count=-1,       // for error message
                  int remove_trailing_space=0 );
			/*
			Read string via "li_get". 
			OBS: Leading white space is not removed.
        		Returns the number of characters read. In case of 
			overflow, an error message is written to cerr and 
			"bufsize" is returned.
			*/


///////////////////////////////////////////////////////////////////

template <class number>
int li_get_unumber( istream& is, number& x )
{ //==============
    number y=0;
    int digits=0;
    char ch;
    li_eatspace(is);
    for (;;) {
        if ( !li_get(is,ch) ) break;
        if ( !isdigit(ch) ) { is.putback(ch); break; }
        digits++;       // OBS: overflow check not implemented
        y = 10*y + (ch-'0');
    }
    if ( digits ) x=y;
    return digits;
}


#endif
